@@ -53,7 +53,7 @@ function _onMenuStateChanged(menu, isOpen) {
 
     let [x, y,] = global.get_pointer();
     let actor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE, x, y);
-    if (Me.stateObj.windowListContains(actor))
+    if (Me.stateObj.someWindowListContains(actor))
         actor.sync_hover();
 }
 
@@ -186,46 +186,41 @@ const WindowTitle = new Lang.Class({
 });
 
 
-const WindowButton = new Lang.Class({
-    Name: 'WindowButton',
+const BaseButton = new Lang.Class({
+    Name: 'BaseButton',
+    Abstract: true,
 
-    _init: function(metaWindow) {
-        this.metaWindow = metaWindow;
+    _init: function(perMonitor, monitorIndex) {
+        this._perMonitor = perMonitor;
+        this._monitorIndex = monitorIndex;
 
-        this._windowTitle = new WindowTitle(this.metaWindow);
         this.actor = new St.Button({ style_class: 'window-button',
                                      x_fill: true,
                                      can_focus: true,
                                      button_mask: St.ButtonMask.ONE |
-                                                  St.ButtonMask.THREE,
-                                     child: this._windowTitle.actor });
+                                                  St.ButtonMask.THREE });
         this.actor._delegate = this;
 
-        this._menuManager = new PopupMenu.PopupMenuManager(this);
-        this._contextMenu = new WindowContextMenu(this.actor, this.metaWindow);
-        this._contextMenu.connect('open-state-changed', _onMenuStateChanged);
-        this._contextMenu.actor.hide();
-        this._menuManager.addMenu(this._contextMenu);
-        Main.uiGroup.add_actor(this._contextMenu.actor);
-
         this.actor.connect('allocation-changed',
                            Lang.bind(this, this._updateIconGeometry));
         this.actor.connect('clicked', Lang.bind(this, this._onClicked));
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
         this.actor.connect('popup-menu', Lang.bind(this, this._onPopupMenu));
 
+        this._contextMenuManager = new PopupMenu.PopupMenuManager(this);
+
         this._switchWorkspaceId =
             global.window_manager.connect('switch-workspace',
                                           Lang.bind(this, this._updateVisibility));
-        this._workspaceChangedId =
-            this.metaWindow.connect('workspace-changed',
-                                    Lang.bind(this, this._updateVisibility));
-        this._updateVisibility();
 
-        this._notifyFocusId =
-            global.display.connect('notify::focus-window',
-                                   Lang.bind(this, this._updateStyle));
-        this._updateStyle();
+        if (this._perMonitor) {
+            this._windowEnteredMonitorId =
+                global.screen.connect('window-entered-monitor',
+                    Lang.bind(this, this._windowEnteredOrLeftMonitor));
+            this._windowLeftMonitorId =
+                global.screen.connect('window-left-monitor',
+                    Lang.bind(this, this._windowEnteredOrLeftMonitor));
+        }
     },
 
     get active() {
@@ -239,6 +234,103 @@ const WindowButton = new Lang.Class({
         this._onClicked(this.actor, 1);
     },
 
+    _onClicked: function(actor, button) {
+        throw new Error('Not implemented');
+    },
+
+    _canOpenPopupMenu: function() {
+        return true;
+    },
+
+    _onPopupMenu: function(actor) {
+        if (!this._canOpenPopupMenu() || this._contextMenu.isOpen)
+            return;
+        _openMenu(this._contextMenu);
+    },
+
+    _isFocused: function() {
+        throw new Error('Not implemented');
+    },
+
+    _updateStyle: function() {
+        if (this._isFocused())
+            this.actor.add_style_class_name('focused');
+        else
+            this.actor.remove_style_class_name('focused');
+    },
+
+    _windowEnteredOrLeftMonitor: function(metaScreen, monitorIndex, metaWindow) {
+        throw new Error('Not implemented');
+    },
+
+    _isWindowVisible: function(window) {
+        let workspace = global.screen.get_active_workspace();
+
+        return !window.skip_taskbar &&
+               window.located_on_workspace(workspace) &&
+               (!this._perMonitor || window.get_monitor() == this._monitorIndex);
+    },
+
+    _updateVisibility: function() {
+        throw new Error('Not implemented');
+    },
+
+    _getIconGeometry: function() {
+        let rect = new Meta.Rectangle();
+
+        [rect.x, rect.y] = this.actor.get_transformed_position();
+        [rect.width, rect.height] = this.actor.get_transformed_size();
+
+        return rect;
+    },
+
+    _updateIconGeometry: function() {
+        throw new Error('Not implemented');
+    },
+
+    _onDestroy: function() {
+        global.window_manager.disconnect(this._switchWorkspaceId);
+
+        if (this._windowEnteredMonitorId)
+            global.screen.disconnect(this._windowEnteredMonitorId);
+        this._windowEnteredMonitorId = 0;
+
+        if (this._windowLeftMonitorId)
+            global.screen.disconnect(this._windowLeftMonitorId);
+        this._windowLeftMonitorId = 0;
+    }
+});
+
+
+const WindowButton = new Lang.Class({
+    Name: 'WindowButton',
+    Extends: BaseButton,
+
+    _init: function(metaWindow, perMonitor, monitorIndex) {
+        this.parent(perMonitor, monitorIndex);
+
+        this.metaWindow = metaWindow;
+        this._updateVisibility();
+
+        this._windowTitle = new WindowTitle(this.metaWindow);
+        this.actor.set_child(this._windowTitle.actor);
+
+        this._contextMenu = new WindowContextMenu(this.actor, this.metaWindow);
+        this._contextMenu.connect('open-state-changed', _onMenuStateChanged);
+        this._contextMenu.actor.hide();
+        this._contextMenuManager.addMenu(this._contextMenu);
+        Main.uiGroup.add_actor(this._contextMenu.actor);
+
+        this._workspaceChangedId =
+            this.metaWindow.connect('workspace-changed',
+                                    Lang.bind(this, this._updateVisibility));
+
+        this._notifyFocusId =
+            global.display.connect('notify::focus-window',
+                                   Lang.bind(this, this._updateStyle));
+        this._updateStyle();
+    },
+
     _onClicked: function(actor, button) {
         if (this._contextMenu.isOpen) {
             this._contextMenu.close();
@@ -251,41 +343,35 @@ const WindowButton = new Lang.Class({
             _openMenu(this._contextMenu);
     },
 
-    _onPopupMenu: function(actor) {
-        if (this._contextMenu.isOpen)
-            return;
-        _openMenu(this._contextMenu);
+    _isFocused: function() {
+        return global.display.focus_window == this.metaWindow;
     },
 
     _updateStyle: function() {
+        this.parent();
+
         if (this.metaWindow.minimized)
             this.actor.add_style_class_name('minimized');
         else
             this.actor.remove_style_class_name('minimized');
+    },
 
-        if (global.display.focus_window == this.metaWindow)
-            this.actor.add_style_class_name('focused');
-        else
-            this.actor.remove_style_class_name('focused');
+    _windowEnteredOrLeftMonitor: function(metaScreen, monitorIndex, metaWindow) {
+        if (monitorIndex == this._monitorIndex && metaWindow == this.metaWindow)
+            this._updateVisibility();
     },
 
     _updateVisibility: function() {
-        let workspace = global.screen.get_active_workspace();
-        this.actor.visible = this.metaWindow.located_on_workspace(workspace);
+        this.actor.visible = this._isWindowVisible(this.metaWindow);
     },
 
     _updateIconGeometry: function() {
-        let rect = new Meta.Rectangle();
-
-        [rect.x, rect.y] = this.actor.get_transformed_position();
-        [rect.width, rect.height] = this.actor.get_transformed_size();
-
-        this.metaWindow.set_icon_geometry(rect);
+        this.metaWindow.set_icon_geometry(this._getIconGeometry());
     },
 
     _onDestroy: function() {
+        this.parent();
         this.metaWindow.disconnect(this._workspaceChangedId);
-        global.window_manager.disconnect(this._switchWorkspaceId);
         global.display.disconnect(this._notifyFocusId);
         this._contextMenu.destroy();
     }
@@ -296,14 +382,14 @@ const AppContextMenu = new Lang.Class({
     Name: 'AppContextMenu',
     Extends: PopupMenu.PopupMenu,
 
-    _init: function(source, app) {
+    _init: function(source, appButton) {
         this.parent(source, 0.5, St.Side.BOTTOM);
 
-        this._app = app;
+        this._appButton = appButton;
 
         this._minimizeItem = new PopupMenu.PopupMenuItem(_("Minimize all"));
         this._minimizeItem.connect('activate', Lang.bind(this, function() {
-            this._getWindowList().forEach(function(w) {
+            this._appButton.getWindowList().forEach(function(w) {
                 w.minimize();
             });
         }));
@@ -311,7 +397,7 @@ const AppContextMenu = new Lang.Class({
 
         this._unminimizeItem = new PopupMenu.PopupMenuItem(_("Unminimize all"));
         this._unminimizeItem.connect('activate', Lang.bind(this, function() {
-            this._getWindowList().forEach(function(w) {
+            this._appButton.getWindowList().forEach(function(w) {
                 w.unminimize();
             });
         }));
@@ -319,7 +405,7 @@ const AppContextMenu = new Lang.Class({
 
         this._maximizeItem = new PopupMenu.PopupMenuItem(_("Maximize all"));
         this._maximizeItem.connect('activate', Lang.bind(this, function() {
-            this._getWindowList().forEach(function(w) {
+            this._appButton.getWindowList().forEach(function(w) {
                 w.maximize(Meta.MaximizeFlags.HORIZONTAL |
                            Meta.MaximizeFlags.VERTICAL);
             });
@@ -328,7 +414,7 @@ const AppContextMenu = new Lang.Class({
 
         this._unmaximizeItem = new PopupMenu.PopupMenuItem(_("Unmaximize all"));
         this._unmaximizeItem.connect('activate', Lang.bind(this, function() {
-            this._getWindowList().forEach(function(w) {
+            this._appButton.getWindowList().forEach(function(w) {
                 w.unmaximize(Meta.MaximizeFlags.HORIZONTAL |
                              Meta.MaximizeFlags.VERTICAL);
             });
@@ -337,22 +423,15 @@ const AppContextMenu = new Lang.Class({
 
         let item = new PopupMenu.PopupMenuItem(_("Close all"));
         item.connect('activate', Lang.bind(this, function() {
-            this._getWindowList().forEach(function(w) {
+            this._appButton.getWindowList().forEach(function(w) {
                 w.delete(global.get_current_time());
             });
         }));
         this.addMenuItem(item);
     },
 
-    _getWindowList: function() {
-        let workspace = global.screen.get_active_workspace();
-        return this._app.get_windows().filter(function(win) {
-            return !win.skip_taskbar && win.located_on_workspace(workspace);
-        });
-    },
-
     open: function(animate) {
-        let windows = this._getWindowList();
+        let windows = this._appButton.getWindowList();
         this._minimizeItem.actor.visible = windows.some(function(w) {
             return !w.minimized;
         });
@@ -372,21 +451,16 @@ const AppContextMenu = new Lang.Class({
 
 const AppButton = new Lang.Class({
     Name: 'AppButton',
+    Extends: BaseButton,
+
+    _init: function(app, perMonitor, monitorIndex) {
+        this.parent(perMonitor, monitorIndex);
 
-    _init: function(app) {
         this.app = app;
+        this._updateVisibility();
 
         let stack = new St.Widget({ layout_manager: new Clutter.BinLayout() });
-        this.actor = new St.Button({ style_class: 'window-button',
-                                     x_fill: true,
-                                     can_focus: true,
-                                     button_mask: St.ButtonMask.ONE |
-                                                  St.ButtonMask.THREE,
-                                     child: stack });
-        this.actor._delegate = this;
-
-        this.actor.connect('allocation-changed',
-                           Lang.bind(this, this._updateIconGeometry));
+        this.actor.set_child(stack);
 
         this._singleWindowTitle = new St.Bin({ x_expand: true,
                                                x_align: St.Align.START });
@@ -409,8 +483,7 @@ const AppButton = new Lang.Class({
         this._menuManager.addMenu(this._menu);
         Main.uiGroup.add_actor(this._menu.actor);
 
-        this._contextMenuManager = new PopupMenu.PopupMenuManager(this);
-        this._appContextMenu = new AppContextMenu(this.actor, this.app);
+        this._appContextMenu = new AppContextMenu(this.actor, this);
         this._appContextMenu.connect('open-state-changed', _onMenuStateChanged);
         this._appContextMenu.actor.hide();
         Main.uiGroup.add_actor(this._appContextMenu.actor);
@@ -421,14 +494,6 @@ const AppButton = new Lang.Class({
                 function() {
                     this._icon.child = app.create_icon_texture(ICON_TEXTURE_SIZE);
                 }));
-        this.actor.connect('clicked', Lang.bind(this, this._onClicked));
-        this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
-        this.actor.connect('popup-menu', Lang.bind(this, this._onPopupMenu));
-
-        this._switchWorkspaceId =
-            global.window_manager.connect('switch-workspace',
-                                          Lang.bind(this, this._updateVisibility));
-        this._updateVisibility();
 
         this._windowsChangedId =
             this.app.connect('windows-changed',
@@ -442,23 +507,30 @@ const AppButton = new Lang.Class({
         this._updateStyle();
     },
 
-    _updateVisibility: function() {
-        let workspace = global.screen.get_active_workspace();
-        this.actor.visible = this.app.is_on_workspace(workspace);
+    _windowEnteredOrLeftMonitor: function(metaScreen, monitorIndex, metaWindow) {
+        if (this._windowTracker.get_window_app(metaWindow) == this.app &&
+            monitorIndex == this._monitorIndex) {
+            this._updateVisibility();
+            this._windowsChanged();
+        }
     },
 
-    _updateStyle: function() {
-        if (this._windowTracker.focus_app == this.app)
-            this.actor.add_style_class_name('focused');
-        else
-            this.actor.remove_style_class_name('focused');
+    _updateVisibility: function() {
+        if (!this._perMonitor) {
+            // fast path: use ShellApp API to avoid iterating over all windows.
+            let workspace = global.screen.get_active_workspace();
+            this.actor.visible = this.app.is_on_workspace(workspace);
+        } else {
+            this.actor.visible = this.getWindowList().length >= 1;
+        }
+    },
+
+    _isFocused: function() {
+        return this._windowTracker.focus_app == this.app;
     },
 
     _updateIconGeometry: function() {
-        let rect = new Meta.Rectangle();
-
-        [rect.x, rect.y] = this.actor.get_transformed_position();
-        [rect.width, rect.height] = this.actor.get_transformed_size();
+        let rect = this._getIconGeometry();
 
         let windows = this.app.get_windows();
         windows.forEach(function(w) {
@@ -467,15 +539,14 @@ const AppButton = new Lang.Class({
     },
 
 
-    _getWindowList: function() {
-        let workspace = global.screen.get_active_workspace();
-        return this.app.get_windows().filter(function(win) {
-            return !win.skip_taskbar && win.located_on_workspace(workspace);
-        });
+    getWindowList: function() {
+        return this.app.get_windows().filter(Lang.bind(this, function(win) {
+            return this._isWindowVisible(win);
+        }));
     },
 
     _windowsChanged: function() {
-        let windows = this._getWindowList();
+        let windows = this.getWindowList();
         this._singleWindowTitle.visible = windows.length == 1;
         this._multiWindowTitle.visible = !this._singleWindowTitle.visible;
 
@@ -507,17 +578,6 @@ const AppButton = new Lang.Class({
 
     },
 
-    get active() {
-        return this.actor.has_style_class_name('focused');
-    },
-
-    activate: function() {
-        if (this.active)
-            return;
-
-        this._onClicked(this.actor, 1);
-    },
-
     _onClicked: function(actor, button) {
         let menuWasOpen = this._menu.isOpen;
         if (menuWasOpen)
@@ -531,7 +591,7 @@ const AppButton = new Lang.Class({
             if (menuWasOpen)
                 return;
 
-            let windows = this._getWindowList();
+            let windows = this.getWindowList();
             if (windows.length == 1) {
                 if (contextMenuWasOpen)
                     return;
@@ -555,20 +615,17 @@ const AppButton = new Lang.Class({
         }
     },
 
-    _onPopupMenu: function(actor) {
-        if (this._menu.isOpen || this._contextMenu.isOpen)
-            return;
-        _openMenu(this._contextMenu);
+    _canOpenPopupMenu: function() {
+        return !this._menu.isOpen;
     },
 
-
     _onMenuActivate: function(menu, child) {
         child._window.activate(global.get_current_time());
     },
 
     _onDestroy: function() {
+        this.parent();
         this._textureCache.disconnect(this._iconThemeChangedId);
-        global.window_manager.disconnect(this._switchWorkspaceId);
         this._windowTracker.disconnect(this._notifyFocusId);
         this.app.disconnect(this._windowsChangedId);
         this._menu.destroy();
@@ -754,7 +811,10 @@ const WorkspaceIndicator = new Lang.Class({
 const WindowList = new Lang.Class({
     Name: 'WindowList',
 
-    _init: function() {
+    _init: function(perMonitor, monitor) {
+        this._perMonitor = perMonitor;
+        this._monitor = monitor;
+
         this.actor = new St.Widget({ name: 'panel',
                                      style_class: 'bottom-panel',
                                      reactive: true,
@@ -788,26 +848,35 @@ const WindowList = new Lang.Class({
         this._workspaceIndicator = new WorkspaceIndicator();
         indicatorsBox.add(this._workspaceIndicator.container, { expand: false, y_fill: true });
 
+        this._workspaceSettings = this._getWorkspaceSettings();
+        this._workspacesOnlyOnPrimaryChangedId =
+            this._workspaceSettings.connect('changed::workspaces-only-on-primary',
+                                            Lang.bind(this, this._updateWorkspaceIndicatorVisibility));
+        this._updateWorkspaceIndicatorVisibility();
+
         this._menuManager = new PopupMenu.PopupMenuManager(this);
         this._menuManager.addMenu(this._workspaceIndicator.menu);
 
-        this._trayButton = new TrayButton();
-        indicatorsBox.add(this._trayButton.actor, { expand: false });
+        this._isOnBottomMonitor = this._monitor == Main.layoutManager.bottomMonitor;
+
+        if (this._isOnBottomMonitor) {
+            this._trayButton = new TrayButton();
+            indicatorsBox.add(this._trayButton.actor, { expand: false });
+        }
 
         Main.layoutManager.addChrome(this.actor, { affectsStruts: true,
                                                    trackFullscreen: true });
         Main.ctrlAltTabManager.addGroup(this.actor, _("Window List"), 'start-here-symbolic');
 
+        this.actor.width = this._monitor.width;
+        this.actor.set_position(this._monitor.x,
+                                this._monitor.y + this._monitor.height - this.actor.height);
+
         this._appSystem = Shell.AppSystem.get_default();
         this._appStateChangedId =
             this._appSystem.connect('app-state-changed',
                                     Lang.bind(this, this._onAppStateChanged));
 
-        this._monitorsChangedId =
-            Main.layoutManager.connect('monitors-changed',
-                                       Lang.bind(this, this._updatePosition));
-        this._updatePosition();
-
         this._keyboardVisiblechangedId =
             Main.layoutManager.connect('keyboard-visible-changed',
                 Lang.bind(this, function(o, state) {
@@ -841,8 +910,6 @@ const WindowList = new Lang.Class({
                 this._updateMessageTrayAnchor();
             }));
 
-        this._isOnBottomMonitor = Main.layoutManager.primaryIndex == Main.layoutManager.bottomIndex;
-
         if (this._isOnBottomMonitor) {
             let actor = this.actor;
             this._bottomHoverChangedId =
@@ -887,6 +954,13 @@ const WindowList = new Lang.Class({
         this._groupingModeChanged();
     },
 
+    _getWorkspaceSettings: function() {
+        let settings = global.get_overrides_settings();
+        if (settings.list_keys().indexOf('workspaces-only-on-primary') > -1)
+            return settings;
+        return new Gio.Settings({ schema_id: 'org.gnome.mutter' });
+    },
+
     _onScrollEvent: function(actor, event) {
         let direction = event.get_scroll_direction();
         let diff = 0;
@@ -912,6 +986,12 @@ const WindowList = new Lang.Class({
         children[active].activate();
     },
 
+    _updateWorkspaceIndicatorVisibility: function() {
+        this._workspaceIndicator.actor.visible =
+            this._monitor == Main.layoutManager.primaryMonitor ||
+            !this._workspaceSettings.get_boolean('workspaces-only-on-primary');
+    },
+
     _getPreferredUngroupedWindowListWidth: function() {
         if (this._windowList.get_n_children() == 0)
             return this._windowList.get_preferred_width(-1)[1];
@@ -921,13 +1001,21 @@ const WindowList = new Lang.Class({
         let spacing = this._windowList.layout_manager.spacing;
 
         let workspace = global.screen.get_active_workspace();
-        let nWindows = global.display.get_tab_list(Meta.TabList.NORMAL, workspace).length;
+        let windows = global.display.get_tab_list(Meta.TabList.NORMAL, workspace);
+        if (this._perMonitor) {
+            windows = windows.filter(Lang.bind(this, function(window) {
+                return window.get_monitor() == this._monitor.index;
+            }));
+        }
+        let nWindows = windows.length;
+        if (nWindows == 0)
+            return this._windowList.get_preferred_width(-1)[1];
 
         return nWindows * childWidth + (nWindows - 1) * spacing;
     },
 
     _getMaxWindowListWidth: function() {
-        let indicatorsBox = this._trayButton.actor.get_parent();
+        let indicatorsBox = this._workspaceIndicator.actor.get_parent();
         return this.actor.width - indicatorsBox.get_preferred_width(-1)[1];
     },
 
@@ -978,12 +1066,6 @@ const WindowList = new Lang.Class({
         }
     },
 
-    _updatePosition: function() {
-        let monitor = Main.layoutManager.primaryMonitor;
-        this.actor.width = monitor.width;
-        this.actor.set_position(monitor.x, monitor.y + monitor.height - this.actor.height);
-    },
-
     _updateKeyboardAnchor: function() {
         if (!Main.keyboard.actor)
             return;
@@ -1013,7 +1095,7 @@ const WindowList = new Lang.Class({
     },
 
     _addApp: function(app) {
-        let button = new AppButton(app);
+        let button = new AppButton(app, this._perMonitor, this._monitor.index);
         this._windowList.layout_manager.pack(button.actor,
                                              true, true, true,
                                              Clutter.BoxAlignment.START,
@@ -1046,7 +1128,7 @@ const WindowList = new Lang.Class({
                 return;
         }
 
-        let button = new WindowButton(win);
+        let button = new WindowButton(win, this._perMonitor, this._monitor.index);
         this._windowList.layout_manager.pack(button.actor,
                                              true, true, true,
                                              Clutter.BoxAlignment.START,
@@ -1155,6 +1237,8 @@ const WindowList = new Lang.Class({
     },
 
     _onDestroy: function() {
+        this._workspaceSettings.disconnect(this._workspacesOnlyOnPrimaryChangedId);
+
         this._workspaceIndicator.destroy();
 
         Main.ctrlAltTabManager.removeGroup(this.actor);
@@ -1162,9 +1246,6 @@ const WindowList = new Lang.Class({
         this._appSystem.disconnect(this._appStateChangedId);
         this._appStateChangedId = 0;
 
-        Main.layoutManager.disconnect(this._monitorsChangedId);
-        this._monitorsChangedId = 0;
-
         Main.layoutManager.disconnect(this._keyboardVisiblechangedId);
         this._keyboardVisiblechangedId = 0;
 
@@ -1211,34 +1292,69 @@ const Extension = new Lang.Class({
     Name: 'Extension',
 
     _init: function() {
-        this._windowList = null;
+        this._windowLists = null;
         this._injections = {};
     },
 
     enable: function() {
-        this._windowList = new WindowList();
+        this._windowLists = [];
 
         this._injections['_trayDwellTimeout'] =
             MessageTray.MessageTray.prototype._trayDwellTimeout;
         MessageTray.MessageTray.prototype._trayDwellTimeout = function() {
             return false;
         };
+
+        this._settings = Convenience.getSettings();
+        this._showOnAllMonitorsChangedId =
+            this._settings.connect('changed::show-on-all-monitors',
+                                   Lang.bind(this, this._buildWindowLists));
+
+        this._monitorsChangedId =
+            Main.layoutManager.connect('monitors-changed',
+                                       Lang.bind(this, this._buildWindowLists));
+
+        this._buildWindowLists();
+    },
+
+    _buildWindowLists: function() {
+        this._windowLists.forEach(function(windowList) {
+            windowList.actor.destroy();
+        });
+        this._windowLists = [];
+
+        let showOnAllMonitors = this._settings.get_boolean('show-on-all-monitors');
+
+        Main.layoutManager.monitors.forEach(Lang.bind(this, function(monitor) {
+            if (showOnAllMonitors || monitor == Main.layoutManager.primaryMonitor)
+                this._windowLists.push(new WindowList(showOnAllMonitors, monitor));
+        }));
     },
 
     disable: function() {
-        if (!this._windowList)
+        if (!this._windowLists)
             return;
 
-        this._windowList.actor.hide();
-        this._windowList.actor.destroy();
-        this._windowList = null;
+        this._settings.disconnect(this._showOnAllMonitorsChangedId);
+        this._showOnAllMonitorsChangedId = 0;
+
+        Main.layoutManager.disconnect(this._monitorsChangedId);
+        this._monitorsChangedId = 0;
+
+        this._windowLists.forEach(function(windowList) {
+            windowList.actor.hide();
+            windowList.actor.destroy();
+        });
+        this._windowLists = null;
 
         for (let prop in this._injections)
             MessageTray.MessageTray.prototype[prop] = this._injections[prop];
     },
 
-    windowListContains: function(actor) {
-        return this._windowList.actor.contains(actor);
+    someWindowListContains: function(actor) {
+        return this._windowLists.some(function(windowList) {
+            return windowList.actor.contains(actor);
+        });
     }
 });
 
